"""
Wrapper class over DB operations
"""
from datetime import datetime

from bson import ObjectId
from geojson import Point

from .database import Database
from .model import User, Taxi, Driver, OperationBoundary, TripInfo, TripInfoDetail
from .taxi_ops_logger import TaxiOpsLogger


class DatabaseManagement:
    """
    Class doc string
    """

    def __init__(self):
        """
        Init Doc string
        """
        self._database = Database()
        self._logger = TaxiOpsLogger()

    def create_user(self, user_name, email_id, phone_number) -> str:
        """
        Makes an entry in User data
        :param user_name:
        :param email_id:
        :param phone_number:
        :return:
        """
        user: User = User(user_name, email_id, phone_number)
        return self._database.insert_single_data(collection="Users",
                                                 data=user.to_dict())  # pylint: disable=no-member

    def create_taxi(self, taxi_type, taxi_number) -> str:
        """
        Makes an entry in Taxis table
        :param taxi_type: taxi type
        :param taxi_number: taxi number
        :return: Object(_id) string
        """
        taxi: Taxi = Taxi(taxi_type=taxi_type, taxi_number=taxi_number, status="Not Operational",
                          location=Point([(0, 0)]))
        return self._database.insert_single_data(collection="Taxis",
                                                 data=taxi.to_dict(),  # pylint: disable=no-member
                                                 index_key="location")

    # driver
    def create_driver(self, driver_name, email_id, phone_number) -> str:
        """
        doc string
        :param driver_name:
        :param email_id:
        :param phone_number:
        :return:
        """
        driver: Driver = Driver(driver_name=driver_name,
                                email_id=email_id,
                                phone_number=phone_number)
        return self._database.insert_single_data(collection="Drivers",
                                                 data=driver.to_dict())  # pylint: disable=no-member

    def get_taxis(self, taxi_filter):
        """
        Gets all the drivers as per filter. In case, all is required, pass empty filter
        @param taxi_filter:
        @return:
        """
        taxi_documents = self._database.find_multiple("Taxis", taxi_filter)
        taxi_list = []
        for taxi_document in taxi_documents:
            taxi = Taxi(taxi_type=taxi_document['taxi_type'],
                        taxi_number=taxi_document['taxi_number'],
                        status=taxi_document['status'],
                        driver_id=taxi_document['driver_id'],
                        location=taxi_document['location'],
                        taxi_id=str(taxi_document['_id']))
            taxi_list.append(taxi)
        return taxi_list

    def get_drivers(self, driver_filter):
        """
        Gets all the drivers as per filter. In case, all is required, pass empty filter
        @param driver_filter:
        @return:
        """
        driver_documents = self._database.find_multiple("Drivers", driver_filter)
        driver_list = []
        for driver_document in driver_documents:
            driver = Driver(driver_name=driver_document['driver_name'],
                            email_id=driver_document['email_id'],
                            phone_number=driver_document['phone_number'],
                            status=driver_document["status"],
                            taxi_assigned=driver_document['taxi_assigned'],
                            location=driver_document['location'],
                            driver_id=str(driver_document['_id']))
            driver_list.append(driver)
        return driver_list

    def register_driver_with_taxi(self, driver: Driver, taxi: Taxi) -> bool:
        """
        update taxi with driver_id and mark driver as taxi_assigned.
        :param driver:
        :param taxi:
        :return:
        """
        matched_count, modified_count = \
            self._database.update_single("Taxis",
                                         {"_id": ObjectId(taxi.taxi_id)},
                                         {"$set": {'driver_id': taxi.driver_id}})
        self._logger.info("Taxi %s record Update: matched: %s and modified: %s",
                          taxi.taxi_number, matched_count, modified_count)
        if matched_count == modified_count == 1:
            matched_count, modified_count = \
                self._database.update_single("Drivers",
                                             {"_id": ObjectId(driver.driver_id)},
                                             {"$set": {'taxi_assigned': driver.taxi_assigned}})
            self._logger.info("Driver %s record Update: matched: %s and modified: %s",
                              driver.driver_name, matched_count, modified_count)

        self._logger.info("Taxi: %s is allocated to driver %s",
                          taxi.taxi_number, driver.driver_name)
        return matched_count == modified_count == 1
        # self._database.run_as_transactions(func_list)

    def get_taxi_current_status(self, taxi_id):
        """
        doc string
        :param taxi_id:
        :return:
        """
        taxi = self._database.find_one("Taxis", {"_id": ObjectId(taxi_id)})
        return taxi['status']

    def update_taxi_status(self, taxi_id, status) -> bool:
        """
        doc string
        :param taxi_id:
        :param status:
        :return:
        """
        matched_count, modified_count = self._database \
            .update_single(collection="Taxis",
                           update_filter={"_id": ObjectId(taxi_id)},
                           update={"$set": {'status': status}})
        return matched_count == modified_count == 1

    def update_driver_status(self, driver_id, status) -> bool:
        """
        doc string
        :param driver_id:
        :param status:
        :return:
        """
        matched_count, modified_count = self._database \
            .update_single(collection="Drivers",
                           update_filter={"_id": ObjectId(driver_id)},
                           update={"$set": {'status': status}})
        return matched_count == modified_count == 1

    def get_current_driver(self, taxi_id):
        """
        doc string
        :param taxi_id:
        :return:
        """

    def get_user_booking(self, user_id):
        """
        doc string
        :param user_id:
        :return:
        """

    def create_operation_bound(self, city,
                               left_bottom: (float, float),
                               left_top: (float, float),
                               right_top: (float, float),
                               right_bottom: (float, float)):
        """
        create operation boundary for city
        @param city:
        @param left_bottom:
        @param left_top:
        @param right_top:
        @param right_bottom:
        @return:
        """
        polygon = [left_bottom, left_top, right_top, right_bottom]
        operational_boundary = OperationBoundary(city, polygon)
        self._logger.info("Inserting Operational Boundary - %s",
                          operational_boundary.to_dict())  # pylint: disable=no-member
        return self._database.insert_single_data("Operation_Boundary",
                                                 operational_boundary.to_dict())  # pylint: disable=no-member

    def get_operation_bound(self, city):
        """
        get operation boundary
        @param city:
        @return:
        """
        document = self._database.find_one("Operation_Boundary", {"city": city})
        polygon_coordinates = document["location"]
        self._logger.info("Fetched polygon for %s - %s", city, polygon_coordinates)
        return polygon_coordinates

    def update_taxi_location(self, taxi_id, location) -> bool:
        """
        update taxi location
        @param taxi_id:
        @param location:
        @return:
        """
        matched_count, modified_count = self._database \
            .update_single(collection="Taxis",
                           update_filter={"_id": ObjectId(taxi_id)},
                           update={"$set": {
                               'location': location["coordinates"]
                           }
                           })
        return matched_count == modified_count == 1

    def find_nearest_taxi(self, current_location,
                          max_distance, max_number_of_taxi) -> list:
        """
        find the nearest taxi
        @param current_location:
        @param max_distance:
        @param max_number_of_taxi:
        @return:
        """
        longitude = current_location["coordinates"][0]
        latitude = current_location["coordinates"][1]
        documents = self._database \
            .find_nearest_entities_in_collection("Taxis",
                                                 longitude,
                                                 latitude,
                                                 max_distance,
                                                 max_number_of_taxi)
        taxi_list = []
        for taxi_document in documents:
            taxi = Taxi(taxi_type=taxi_document['taxi_type'],
                        taxi_number=taxi_document['taxi_number'],
                        status=taxi_document['status'],
                        driver_id=taxi_document['driver_id'],
                        location=taxi_document['location'],
                        taxi_id=str(taxi_document['_id']))
            taxi_list.append(taxi)

        return taxi_list

    def create_trip(self, taxi_id, user_id, current_location, destination_location) -> str:
        """
        create trip
        @param taxi_id:
        @param user_id:
        @param current_location:
        @param destination_location:
        @return:
        """
        trip_info = TripInfo(trip_id="", user_id=user_id, taxi_id=taxi_id,
                             starting_point=current_location,
                             destination_point=destination_location,
                             status="Not Started")
        self._logger.info("Now creating trip with details - ")
        self._logger.info(trip_info.to_dict())  # pylint: disable=no-member
        return self._database.insert_single_data("TripInfos", trip_info.to_dict())  # pylint: disable=no-member

    def mark_trip_in_progress(self, trip_id) -> bool:
        """

        @param trip_id:
        @return:
        """
        matched_count, modified_count = self._database \
            .update_single(collection="TripInfos",
                           update_filter={"_id": ObjectId(trip_id)},
                           update={"$set": {
                               "status": "In Progress",
                               "start_time": datetime.now()}})
        return matched_count == modified_count == 1

    def insert_trip_location(self, trip_id, location, eta):
        """
        insert new location
        @param trip_id:
        @param location:
        @param eta:
        @return:
        """
        trip_info_detail = TripInfoDetail(trip_id=trip_id,
                                          location=location,
                                          reporting_time=datetime.now(),
                                          expected_time_for_completion_in_min=eta)
        return self._database.insert_single_data(collection="TripInfoDetail",
                                                 data=trip_info_detail.to_dict())

    def clean_up_database(self):
        """clean up database"""
        for collection in ["Taxis", "Users", "Drivers", "Operation_Boundary",
                           "TripInfos", "TripInfoDetail"]:
            self._database.drop_collection(collection=collection)
        # for collection in ["Operation_Boundary"]:
        #     self._database.drop_collection(collection=collection)



"""
DB operations
"""
from typing import Mapping, Any, Union

import pymongo
from bson import SON
# Imports MongoClient for base level access to the local MongoDB
from pymongo import MongoClient
from pymongo.errors import ConfigurationError
from pymongo.typings import _Pipeline

from .taxi_ops_logger import TaxiOpsLogger


class Database:
    """
    Class for DB operations
    """
    # Class variables used for database host ip and port information, database name
    # variables are referred to by using <class_name>.<variable_name>
    # AK_URI = "mongodb+srv://apache_abhinav_official:6g6RRf2UZiJy8G2" \
    #          "@taxiops.jdfqcmn.mongodb.net/?retryWrites=true&w=majority"
    # AK_DB_NAME = 'TaxiOps'
    # _instance = None initially
    # _db_conn = database connection client
    # _db = database
    # _logger = applevel logger

    DB_URI = "mongodb+srv://iitmcloudfeb22g1:iitmcloudfeb22g1@spn.xqzqo3y.mongodb.net/" \
             "?retryWrites=true&w=majority"
    DB_NAME = 'ProjectTaxi'
    _instance = None
    _db_conn = MongoClient(DB_URI)
    _db = _db_conn[DB_NAME]
    _logger = TaxiOpsLogger()

    def __new__(cls):
        """
        making database object singleton class
        :return: Database instance
        """
        if cls._instance is None:
            cls._logger.info('Creating the database instance')
            cls._instance = super(Database, cls).__new__(cls)
        return cls._instance

    def find_one(self, collection, key):
        """
        This method finds a single document using field information provided in the key parameter.
        It assumes that the key returns a unique document. It returns None if no document is found
        :param collection:
        :param key:
        :return:
        """
        db_collection = self._db[collection]
        document = db_collection.find_one(key)
        return document

    def find_multiple(self, collection, data):
        """
        This method finds multiple documents using field information provided in the key parameter.
        It returns cursor on valid query fetch. Here we have suppressed _id being sent out
        :param collection:
        :param data:
        :return:
        """
        try:
            if data is None:
                self._logger.warning("No data is available. Please provide the data.")
                return True
            db_collection = self._db[collection]
            result = db_collection.find(data)
            return result
        except TypeError as exception:
            self._logger.error("An exception occurred :: %s", exception)
            return False

    def insert_single_data(self, collection: str, data: dict, index_key: str = "") -> str:
        """ This method inserts the data in a new document. It assumes that any uniqueness check
        is done by the caller. index_key is currently supported for geo_spatial only in this
        project, and it should only once. It is also assumed that developers will use insert single
         or multiple insert, which in turn creates the collection.
        :param collection:
        :param data:
        :param index_key:
        :return:
        """
        db_collection = self._db[collection]
        document = db_collection.insert_one(data)
        # The pymongo.collection.Collection.create_index method only creates an index
        # if an index of the same specification does not already exist.
        # https://www.mongodb.com/docs/manual/indexes/
        if index_key != "":
            self.__create_geo_spatial_index(self._db[collection], index_key)
        return document.inserted_id

    def insert_multiple(self, collection, data, index_key: str = ""):
        """
        This method inserts data. Created by Abhinav
        :param collection:
        :param data:
        :param index_key:
        :return:
        """
        try:
            if data is None:
                self._logger.warning("No data is available.")
                return True
            result = self._db[collection].insert_many(data)
            if index_key != "":
                self.__create_geo_spatial_index(collection, index_key)
            self._logger.info(result.inserted_ids)
            return True
        except TypeError as exception:
            self._logger.error("An exception occurred :: %s", exception)
            return False

    def delete_single(self, collection, data):
        """ This method deletes the data from collection
        :param collection:
        :param data:
        :return:
        """
        if data is not None:
            result = self._db[collection].delete_one(data)
            return result
        self._logger.warning("No data is available. Please provide the data.")
        return True

    def delete_multiple(self, collection, data):
        """ This method deletes the data from collection
        :param collection:
        :param data:
        :return:
        """
        if data is not None:
            result = self._db[collection].delete_many(data)
            return result
        self._logger.warning("No data is available. Please provide the data.")
        return True

    def update_single(self, collection, update_filter: Mapping[str, Any],
                      update: Union[Mapping[str, Any], _Pipeline]):
        """
        This method update the data from Collection
        {{{
        collection = db.[collection]

        # Updating fan quantity from 10 to 25.
        filter = { 'appliance': 'fan' }

        # Values to be updated.
        new_values = { "$set": { 'quantity': 25 } }

        collection.update_one(filter, new_values)
        }}}
        :param collection: The Collection on which the operation needs to run
        :param update_filter: filter to be used
        :param update: data to be updated
        :return:
        """
        if update_filter is not None and update is not None:
            result = self._db[collection].update_one(filter=update_filter, update=update)
            return result.matched_count, result.modified_count
        self._logger.warning("No data is available. Please provide the data.")
        return None, None

    def update_multiple(self, collection, update_filter: Mapping[str, Any],
                        update: Union[Mapping[str, Any], _Pipeline], upsert: bool = False):
        """ This method deletes the data from collection
        :param collection:
        :param update_filter:
        :param update:
        :param upsert:
        :return:
        """
        if update_filter is None or update is None:
            self._logger.warning("No data is available. Please provide the data.")
            return True
        return self._db[collection].update_many(filter=update_filter, update=update, upsert=upsert)

    def run_as_transactions(self, func_list: []):
        """Work in Progress. Transactions"""
        with self._db_conn.start_session() as session:
            with session.start_transaction():
                for func in func_list:
                    func(session)

    def aggregate_data(self, collection, query):
        """
        This method perform aggregation on collection. Created by Abhinav
        :param collection:
        :param query:
        :return:
        """
        result = self._db[collection].aggregate(query)
        return result

    def drop_collection(self, collection):
        """
        This is only created to support testing
        :param collection:
        :return:
        """
        try:
            self._db[collection].drop()
            self._logger.info("Collection: %s dropped", collection)
            return True
        except ConfigurationError as exception:
            self._logger.error("An exception occurred :: %s", exception)
            return False

    def count_collection(self, collection, key):
        """
        This method returns counts of document for the filter in collection
        :param collection:
        :param key:
        :return:
        """
        result = self._db[collection].count(key)
        return result

    def find_nearest_entities_in_collection(self, collection,
                                            lon, lat, max_distance, max_number_of_taxi):
        """
        Finding the nearest entities in collection
        @param collection:
        @param lon:
        @param lat:
        @param max_distance:
        @param max_number_of_taxi:
        @return:
        """

        try:
            db_collection = self._db[collection]
            query = {"location": SON([("$nearSphere", [lon, lat]),
                                      ("$maxDistance", max_distance)]
                                     )
                     }
            documents_near = db_collection.find(query).limit(max_number_of_taxi)
            document_list = []
            for doc in documents_near:
                document_list.append(doc)
            return document_list
        except Exception as exception:
            self._logger.error('Exception in geoNear: %s', exception)
            raise

    def __create_geo_spatial_index(self, collection, key: str):
        try:
            self._db[collection].create_index([(key, pymongo.GEOSPHERE)])
            self._logger.info("Geospatial index created on collection - %s, key - %s",
                              collection, key)
        except TypeError as exception:
            self._logger.error("An exception occurred :: %s", exception)
        except ConfigurationError as exception:
            self._logger.error("An exception occurred :: %s", exception)



"""Data model to used in this project"""

import datetime
from geojson import Point


# @dataclass
class User:  # pylint: disable=too-few-public-methods
    """User Table"""
    # attributes
    name: str
    email_id: str
    phone_number: str
    location: Point
    # will work out: user_id remains blank in DB
    user_id: str

    def __init__(self, name, email_id, phone_number, location=None, user_id=''):
        self.name = name
        self.email_id = email_id
        self.phone_number = phone_number
        self.location = location
        self.user_id = user_id

    def to_dict(self):
        """convert to dict"""
        return {"name": self.name, "email_id": self.email_id,
                "phone_number": self.phone_number,
                "location": "", "user_id": self.user_id}


# @dataclass
class Driver:  # pylint: disable=too-few-public-methods
    """Driver Table"""
    driver_name: str
    email_id: str
    phone_number: str
    status: str
    taxi_assigned: bool
    location: Point
    # will work out: driver_id remains blank in DB
    driver_id: str

    def __init__(self, driver_name: str,
                 email_id: str,
                 phone_number: str,
                 status: str = "OFF DUTY",
                 taxi_assigned: bool = False,
                 location: Point = None,
                 # will work out: driver_id remains blank in DB
                 driver_id: str = ""):
        self.driver_name = driver_name
        self.email_id = email_id
        self.phone_number = phone_number
        self.status = status
        self.taxi_assigned = taxi_assigned
        self.location = location
        self.driver_id = driver_id

    def to_dict(self):
        """convert to dict"""
        return {"driver_name": self.driver_name, "email_id": self.email_id,
                "phone_number": self.phone_number, "status": self.status,
                "taxi_assigned": self.taxi_assigned, "driver_id": self.driver_id,
                "location": ""}


TaxiType = ["Utility", "Deluxe", "Luxury"]
TaxiStatus = ["Not Operational", "Occupied", "Available"]


# @dataclass
class Taxi:  # pylint: disable=too-few-public-methods
    """Taxi table"""
    taxi_type: str
    taxi_number: str
    status: str
    location: Point
    # Assumption for this project is that there would be 1 driver for a taxi
    driver_id: str
    # will work out: taxi_id remains blank in DB
    taxi_id: str

    def __init__(self, taxi_type, taxi_number, status, location, driver_id='', taxi_id=''):
        self.taxi_type = taxi_type
        self.taxi_number = taxi_number
        self.status = status
        self.location = location
        self.driver_id = driver_id
        self.taxi_id = taxi_id

    def to_dict(self):
        """converts taxi object to dict"""
        geojson_point = None
        if self.location is not None:
            # geojson_point = {"type": "Point", "coordinates": self.location["coordinates"][0]}
            geojson_point = self.location["coordinates"][0]

        return {"taxi_type": self.taxi_type, "taxi_number": self.taxi_number,
                "status": self.status, "driver_id": self.driver_id,
                "location": geojson_point}


# @dataclass
class OperationBoundary:  # pylint: disable=too-few-public-methods
    """Operation Boundary"""
    city: str
    location: list[(float, float)]

    def __init__(self, city, location):
        self.city = city
        self.location = location

    def to_dict(self):
        """convert to dict"""
        return {"city": self.city, "location": self.location}


TripStatus = ["Not Started", "In Progress", "Completed"]


# @dataclass
class TripInfo:
    """TripInfo Table"""
    trip_id: str
    user_id: str
    taxi_id: str
    starting_point: Point
    destination_point: Point
    status: TripStatus
    start_time: datetime.datetime
    end_time: datetime.datetime

    def __init__(self, trip_id: str,
                 user_id: str,
                 taxi_id: str,
                 starting_point: Point,
                 destination_point: Point,
                 status: TripStatus,
                 start_time: datetime.datetime = None,
                 end_time: datetime.datetime = None):
        self.trip_id = trip_id
        self.user_id = user_id
        self.taxi_id = taxi_id
        self.starting_point = starting_point
        self.destination_point = destination_point
        self.status = status
        self.start_time = start_time
        self.end_time = end_time

    @staticmethod
    def point_to_location(point):
        """
        geojson point to mongodb compatible location
        @param point:
        @return:
        """
        geojson_point = None
        if point is not None:
            geojson_point = point["coordinates"]
        return geojson_point

    def to_dict(self):
        """converts trip info into dict"""
        return {"trip_id": self.trip_id, "user_id": self.user_id,
                "taxi_id": self.taxi_id,
                "starting_point": self.point_to_location(self.starting_point),
                "destination_point": self.point_to_location(self.destination_point),
                "status": self.status, "start_time": self.start_time,
                "end_time": self.start_time}


# @dataclass
class TripInfoDetail:  # pylint: disable=too-few-public-methods
    """TripInfoDetail Table"""
    trip_id: str
    location: Point
    reporting_time: datetime.datetime
    expected_time_for_completion_in_min: int

    def __init__(self, trip_id: str,
                 location: Point,
                 reporting_time: datetime.datetime,
                 expected_time_for_completion_in_min: int):
        self.trip_id = trip_id
        self.location = location
        self.reporting_time = reporting_time
        self.expected_time_for_completion_in_min = expected_time_for_completion_in_min

    def to_dict(self):
        """converts taxi object to dict"""
        geojson_point = None
        if self.location is not None:
            geojson_point = self.location["coordinates"]

        return {"trip_id": self.trip_id, "location": geojson_point,
                "reporting_time": self.reporting_time,
                "expected_time_for_completion_in_min":
                    self.expected_time_for_completion_in_min}



"""
Logging module to be used in the project
"""

import logging


class TaxiOpsLogger(logging.Logger):
    """ Overriden Logger class to be used in the project """
    class TaxiOpsLogStreamHandler(logging.StreamHandler):
        """ Overriden StreamHandler class to be used in the project """
        def __init__(self):
            logging.StreamHandler.__init__(self)
            log_format = '%(asctime)s module:%(module)-15s func:%(funcName)s' \
                         ' line:%(lineno)d - %(message)s'
            fmt_date = '%Y-%m-%dT%T%Z'
            formatter = logging.Formatter(log_format, fmt_date)
            self.setFormatter(formatter)

    class TaxiOpsLogFileHandler(logging.FileHandler):
        """ Overriden FileHandler class to be used in the project """
        def __init__(self, filename):
            logging.FileHandler.__init__(self, filename)
            log_format = '%(asctime)s module:%(module)-15s func:%(funcName)s' \
                         ' line:%(lineno)d - %(message)s'
            fmt_date = '%Y-%m-%dT%T%Z'
            formatter = logging.Formatter(log_format, fmt_date)
            self.setFormatter(formatter)

    def __init__(self):
        logging.Logger.__init__(self, 'TaxiOps', 'DEBUG')
        self.addHandler(TaxiOpsLogger.TaxiOpsLogStreamHandler())



"""
Taxi Administration. This is to be run on EC2 machine because this is
helper to the taxi ops
"""
import random

from .custom_error import IllegalTaxiStatusState
from .database_management import DatabaseManagement
from .taxi_ops_logger import TaxiOpsLogger
from .model import Taxi


def register_taxi(taxi_type, taxi_number):
    """
    doc string
    :param taxi_type:
    :param taxi_number:
    :return:
    """
    return DatabaseManagement().create_taxi(taxi_type=taxi_type, taxi_number=taxi_number)


def register_driver(driver_name, email_id, phone_number) -> str:
    """
    doc string
    :param driver_name:
    :param email_id:
    :param phone_number:
    :return:
    """
    return DatabaseManagement().create_driver(driver_name=driver_name,
                                              email_id=email_id,
                                              phone_number=phone_number)


def register_drivers_with_taxis():
    """registers driver with taxi
    :return: None
    """
    logger = TaxiOpsLogger()
    # get the required list of taxi and driver.
    unallocated_taxis = DatabaseManagement().get_taxis({"driver_id": ""})
    unallocated_drivers = DatabaseManagement().get_drivers({"taxi_assigned": False})

    # just randomise to be fair
    random.shuffle(unallocated_taxis)
    random.shuffle(unallocated_drivers)

    # now assign
    while len(unallocated_drivers) > 0 and len(unallocated_taxis) > 0:
        taxi = unallocated_taxis.pop(0)
        driver = unallocated_drivers.pop(0)
        driver.taxi_assigned = True
        taxi.driver_id = driver.driver_id
        logger.info("Now registering taxi: %s with driver: %s",
                    taxi.taxi_number,
                    driver.driver_name)

        success = False
        retry = 0
        while not success and retry < 3:
            success = DatabaseManagement().register_driver_with_taxi(driver=driver,
                                                                     taxi=taxi)
            retry += 1
        logger.info("Registration Result: %s", success)


def set_taxi_available(taxi_id) -> bool:
    """get available taxis"""
    return DatabaseManagement().update_taxi_status(taxi_id, "Available")


def get_available_taxis() -> list[Taxi]:
    """get available taxis"""
    return DatabaseManagement().get_taxis({"status": "Available"})


def get_taxi_current_status(taxi_id) -> str:
    """get taxi current status"""
    return DatabaseManagement().get_taxi_current_status(taxi_id=taxi_id)


def set_taxis_drivers_ready():
    """Per design, Driver make taxi available when he is ON-DUTY. Taxi should always emit
     location when it is available.
     This function is a series of helper procedure:
         1. The driver gets active - his status gets to ON DUTY
         2. The car becomes available and starts emitting its locations"""
    logger = TaxiOpsLogger()
    driver_list = DatabaseManagement().get_drivers(({}))
    taxi_list = DatabaseManagement().get_taxis(({}))
    taxi_dict: dict[str, Taxi] = {}
    for taxi in taxi_list:
        taxi_dict[taxi.driver_id] = taxi

    for driver in driver_list:
        b_success = DatabaseManagement().update_driver_status(driver_id=driver.driver_id,
                                                              status="ON-DUTY")
        if b_success:
            taxi = taxi_dict[driver.driver_id]
            logger.info("Driver - %s tagged with taxi %s status ON-DUTY: %s",
                        driver.driver_name, taxi.taxi_number, str(b_success))
            b_success = DatabaseManagement().update_taxi_status(taxi_id=taxi.taxi_id,
                                                                status="Available")
            logger.info("Associated taxi - %s status Available: %s", taxi.taxi_number,
                        str(b_success))

    logger.info("***All taxis and driver ready**** Refreshed list")
    taxi_list = DatabaseManagement().get_taxis(({}))
    logger.info(taxi_list)


def find_nearest_taxi(current_location, max_distance, max_number_of_taxi) -> list:
    """
    find the nearest taxi
    @param current_location:
    @param max_distance:
    @param max_number_of_taxi:
    @return:
    """
    return DatabaseManagement().find_nearest_taxi(current_location,
                                                  max_distance,
                                                  max_number_of_taxi)


def update_taxi_location(v_taxi_id, v_location):
    """
    update taxi location
    @param v_taxi_id:
    @param v_location:
    @return:
    """
    logger = TaxiOpsLogger()
    b_success = DatabaseManagement().update_taxi_location(v_taxi_id, v_location)
    logger.info("Taxi movement to %s - completed", v_location)
    return b_success


def book_and_initiate_trip(taxi_id, user_id, current_location, destination_location):
    """
    book_and_initiate_trip
    @param taxi_id:
    @param user_id:
    @param current_location:
    @param destination_location:
    @return:
    """
    # mark taxi_id status and Occupied.
    # Check if taxi_is current state is Available, If so, mark it Occupied. If not, send error
    logger = TaxiOpsLogger()
    current_status = DatabaseManagement().get_taxi_current_status(taxi_id)
    if current_status != "Available":
        raise IllegalTaxiStatusState(f"Taxi has already become {current_status}")

    b_success = DatabaseManagement().update_taxi_status(taxi_id, "Occupied")
    if b_success:
        trip_id = DatabaseManagement().create_trip(taxi_id, user_id,
                                                   current_location, destination_location)
        logger.info("Trip created - %s", trip_id)
        logger.info("Taxi moving to user location...")
        DatabaseManagement().update_taxi_location(taxi_id, current_location)
        logger.info("Taxi movement to %s - completed", current_location)

        b_success = DatabaseManagement().mark_trip_in_progress(trip_id=trip_id)
        if b_success:
            logger.info("Marking trip - %s in progress", trip_id)
            return trip_id, b_success
        return trip_id, b_success
    return None, False


def insert_trip_info_detail(trip_id, location, expected_time_for_completion):
    """
    insert trip info detail
    @param trip_id:
    @param location:
    @param expected_time_for_completion:
    @return:
    """
    return DatabaseManagement() \
        .insert_trip_location(trip_id=trip_id,
                              location=location,
                              eta=expected_time_for_completion)



"""This is the service layer which is exposed using API gateway.
This should go in lambda
"""
from .database_management import DatabaseManagement


def register_user(name, email_id, phone_number) -> str:
    """
    doc string
    :param name:
    :param email_id:
    :param phone_number:
    :return:
    """
    return DatabaseManagement().create_user(user_name=name,
                                            email_id=email_id,
                                            phone_number=phone_number)



"""Module for exception class"""
from .taxi_ops_logger import TaxiOpsLogger


class IllegalTaxiStatusState(Exception):
    """Error Raised when Taxi status is not correct for an action to happen"""
    def __init__(self, error_string):
        TaxiOpsLogger().error(error_string)
        super().__init__(error_string)



"""setup for this project"""
import os
from setuptools import setup, find_packages

current_directory = os.path.dirname(os.path.abspath(__file__))
print(f"current directory is {current_directory}")
try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        LONG_DESCRIPTION = f.read()
        print(LONG_DESCRIPTION)
except FileNotFoundError:
    LONG_DESCRIPTION = ''

packages_to_install = find_packages()
print(packages_to_install)

setup(
    name='taxi_ops_backend',
    version='0.1.5',
    packages=packages_to_install,
    url='https://github.com/abhinavofficial/IITM-ACSEFeb22-Capstone-TaxiOps',
    license='',
    author='Abhinav Kumar',
    author_email='apache.abhinav.official@outlook.com',
    description='Taxi Ops Backend for capstone project',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
    ],
    python_requires='>=3.9',
    # package_dir={'': 'taxi_ops_backend'},
    install_requires=[
        "pymongo==4.3.3",
        "geojson~=3.0.0"
    ]
)



"""This is the service layer which is exposed using API gateway.
This should go in lambda
"""
from .database_management import DatabaseManagement


def define_boundary_location(city: str,
                             left_bottom: (float, float),
                             left_top: (float, float),
                             right_top: (float, float),
                             right_bottom: (float, float)):
    """
    define city operation boundary
    @param city:
    @param left_bottom:
    @param left_top:
    @param right_top:
    @param right_bottom:
    @return:
    """
    return DatabaseManagement().create_operation_bound(city, left_bottom,
                                                       left_top,
                                                       right_top,
                                                       right_bottom)


def get_operation_boundary(city):
    """
    get operation boundary polygon
    @param city:
    @return:
    """
    return DatabaseManagement().get_operation_bound(city=city)
